/*
 * Copyright 2011 HoldenC
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/* fdm.c : This program implements the finite difference method
 * in order to solve simple problems such as finding the electric
 * potential in the region between two perfect conductors.
 *
 * Compile: gcc -Wall -O2 -lm fdm.c -o fdm
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define FLAG_INNER 0x01
#define FLAG_OUTER 0x02
#define FLAG_BOUND 0x04
#define FLAG_BOUND_DIR 0x08
#define FLAG_BOUND_NEU 0x10

/* struct node represent the node of our grid */
struct node {
	int    x;
	int    y;
	int    index;   /* absolute index */
	int    uindex;  /* index of unknown */
	double v;       /* scalar value */
	double bv;      /* boundary value in case of Dirichlet condition */
	int   flags;
};

/* struct equation holds one equation (5 coeff and b) */
struct equation {
	int   coeff[5];
	int   index[5];
	double b;
};

/* domain dimensions, in number of points. Points are numbered from 0 to X-1,
   and from 0 to Y-1. The origin is in the upper left corner, x axis is
   oriented from left to right, and y axis is oriented from top to bottom */
int X;
int Y;

/* converts x,y grid coordinates into absolute index */
int ind(int x, int y)
{
	return (x + y*X);
}

/* returns the absolute index of the north node, or -1 if it doesn't exist */
int north(struct node *n)
{
	int y;

	y = (n->y) - 1;
	if (y < 0)
		return -1;
	return ind(n->x, y);
}

int south(struct node *n)
{
	int y;

	y = (n->y) + 1;
	if (y > Y-1)
		return -1;
	return ind(n->x, y);
}

int east(struct node *n)
{
	int x;

	x = (n->x) + 1;
	if (x > X-1)
		return -1;
	return ind(x, n->y);
}

int west(struct node *n)
{
	int x;

	x = (n->x) - 1;
	if (x < 0)
		return -1;
	return ind(x, n->y);
}

/* checks whether a (x,y) pair is valid */
int checkxy(int x, int y)
{
	if (x < 0 || x >= X || y < 0 || y >= Y) {
		printf("Error: checkxy(%d, %d).\n", x, y);
		return 0;
	}
	return 1;
}

/* mark a boundary line, from (x1,y1) to (x2,y2) */
void mark_boundary_line(struct node *grid, int x1, int y1, int x2, int y2, int type, double bv)
{
	int x, y;
	struct node *pnode;

	if (y1 == y2) {
		y = y1;
		for (x=x1; x<=x2; x++) {
			if (!checkxy(x,y)) continue;
			pnode = grid+ind(x,y);
			pnode->flags &= ~(FLAG_BOUND_DIR | FLAG_BOUND_NEU);
			pnode->flags |= (FLAG_BOUND | type);
			pnode->bv     = bv;
			if (type & FLAG_BOUND_DIR)
				pnode->v  = bv;
		}
	}
	if (x1 == x2) {
		x = x1;
		for (y=y1; y<=y2; y++) {
			if (!checkxy(x,y)) continue;
			pnode = grid+ind(x,y);
			pnode->flags &= ~(FLAG_BOUND_DIR | FLAG_BOUND_NEU);
			pnode->flags |= (FLAG_BOUND | type);
			pnode->bv     = bv;
			if (type & FLAG_BOUND_DIR)
				pnode->v  = bv;
		}
	}
}

/* mark a boundary rectangle given top left and bottom right points */
void mark_boundary_rectangle(struct node *grid, int tlx, int tly, int brx, int bry, int type, double bv)
{
	mark_boundary_line(grid, tlx, tly, brx, tly, type, bv); /* hor top */
	mark_boundary_line(grid, tlx, bry, brx, bry, type, bv); /* hor bot */
	mark_boundary_line(grid, tlx, tly, tlx, bry, type, bv); /* ver left */
	mark_boundary_line(grid, brx, tly, brx, bry, type, bv); /* ver right */
}

/* mark a boundary square given center point and radius */
void mark_boundary_square(struct node *grid, int cx, int cy, int radius, int type, double bv)
{
	mark_boundary_line(grid, cx-radius, cy-radius, cx+radius, cy-radius, type, bv); /* hor top */
	mark_boundary_line(grid, cx-radius, cy+radius, cx+radius, cy+radius, type, bv); /* hor bot */
	mark_boundary_line(grid, cx-radius, cy-radius, cx-radius, cy+radius, type, bv); /* ver left */
	mark_boundary_line(grid, cx+radius, cy-radius, cx+radius, cy+radius, type, bv); /* ver right */
}

/* mark a boundary circle given center point and radius */
void mark_boundary_circle(struct node *grid, int cx, int cy, int radius, int type, double bv)
{
	int x, y;
	double deg, theta;

	for (deg = 0.0; deg < 360.0; deg += 0.5) {
		theta = deg * M_PI / 180.0;
		x = cx + rint(radius * cosf(theta));
		y = cy + rint(radius * sinf(theta));

		if (!checkxy(x,y)) continue;
		(grid+ind(x,y))->flags |= (FLAG_BOUND | type);
		(grid+ind(x,y))->bv     = bv;
		if (type & FLAG_BOUND_DIR)
			(grid+ind(x,y))->v  = bv;
	}
}

/* recursively fill outer region, bounded by boundary points or grid
 * edges, given a point inside the outer region */
void recursive_fill_outer(struct node *grid, int x, int y, double v)
{
	struct node *pnode, *pn;
	int nn, ns, ne, nw;

	if (!checkxy(x,y))
		exit(1);

	pnode = grid+ind(x,y);
	if (pnode->flags & FLAG_OUTER || pnode->flags & FLAG_BOUND)
		return;
	pnode->flags |= FLAG_OUTER;
	pnode->v      = v;

	nn = north(pnode);
	ns = south(pnode);
	ne =  east(pnode);
	nw =  west(pnode);

	if (nn != -1) {
		pn = grid+nn;
		recursive_fill_outer(grid, pn->x, pn->y, v);
	}

	if (ns != -1) {
		pn = grid+ns;
		recursive_fill_outer(grid, pn->x, pn->y, v);
	}

	if (ne != -1) {
		pn = grid+ne;
		recursive_fill_outer(grid, pn->x, pn->y, v);
	}

	if (nw != -1) {
		pn = grid+nw;
		recursive_fill_outer(grid, pn->x, pn->y, v);
	}
}

/* print ascii representation of the grid */
void printgrid(struct node *grid)
{
	int i;
	struct node *pnode;

	for (i=0; i<X*Y; i++) {
		pnode = grid+i;

		if (pnode->flags & FLAG_INNER)
			printf("· ");
			//printf("%.0f ", pnode->v);
			//printf("%02d_%02d ", pnode->y, pnode->x);
		if (pnode->flags & FLAG_OUTER)
			printf("  ");
		if (pnode->flags & FLAG_BOUND) {
			if (pnode->flags & FLAG_BOUND_DIR)
				//printf("%.0f ", pnode->bv);
				printf("D ");
			if (pnode->flags & FLAG_BOUND_NEU)
				printf("N ");
		}

		if (pnode->x == X-1)
			printf("\n");
	}
}

void usage(void)
{
	printf("usage: fdm -p problem -x xn [-y yn] <outfile>\n");
	printf("       -p problem  problem to solve\n");
	printf("       -x xn       number of horizontal points\n");
	printf("       -y yn       number of vertical points (xn if not specified)\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	int i, j, ii, N;
	int nn, ns, ne, nw;
	int ninner, nouter, nbound, nbound_dir, nbound_neu;
	int problem, opt;
	double k;
	struct node *grid, *pnode, *pn;
	int *uind;                 /* indices of unknowns */
	struct equation *AB, *peq; /* system of equations */
	double *Xold, *Xnew, *aux, rms, rmsnew;  /* */
	double cnorm, dx, dy;
	FILE *fp;
	char filename[256];

	/* read command line parameters */
	if (argc < 2)
		usage();

	problem = 1;
	X = -1;
	Y = -1;
	while ((opt = getopt(argc, argv, "p:x:y:")) != -1) {
		switch (opt) {
			case 'p':
				problem = atoi(optarg);
				break;
			case 'x':
				X = atoi(optarg);
				break;
			case 'y':
				Y = atoi(optarg);
				break;
			default:
				usage();
		}
	}
	if (X == -1) X = 50;
	if (Y == -1) Y = X;
	if (X < 0 || Y < 0 || X*Y > 400*400) {
		printf("Error on parameters: X=%d, Y=%d\n", X, Y);
		exit(1);
	}
	if (problem < 1 || problem > 6) {
			printf("Error: problem not defined\n");
			exit(1);
	}

	//printf("argc = %d, optind = %d\n", argc, optind);

	/* read output file name */
	if (argc == optind) {
		usage();
	} else {
		if (strlen(argv[optind]) > 255) {
			printf("Error: output file name too long\n");
			exit(1);
		}
		else
			strcpy(filename, argv[optind]);
	}

	/* open output file */		
	fp = fopen(filename, "w");
	if (fp == NULL) {
		printf("Error: cannot create file %s\n", filename);
		exit(1);
	}

	/* grid memory allocation */
	grid = malloc(X * Y * sizeof(struct node));
	if (!grid) {
		printf("Error on malloc()\n");
		exit(1);
	}

	/* grid setup*/
	for (i=0; i<X*Y; i++) {
		pnode = grid+i;

		pnode->x      = i % X;
		pnode->y      = i / X;
		pnode->index  = i;
		pnode->uindex = -1;
		pnode->v      = 0.0;
		pnode->bv     = 0.0;
		pnode->flags  = 0;
	}

	/* problem and boundary setup */

	switch (problem) {
		case 1:
		/* one conductor, 1/4 domain */
		k = 0.7;
		/* mark dirichlet boundary */
		mark_boundary_line(grid,           0,             0,         X-1,             0, FLAG_BOUND_DIR, 0); /* hor top */
		mark_boundary_line(grid,         X-1,             0,         X-1,           Y-1, FLAG_BOUND_DIR, 0); /* ver right*/
		mark_boundary_line(grid,           0, rint((1-k)*Y), rint(k*X-1), rint((1-k)*Y), FLAG_BOUND_DIR, 1); /* hor */
		mark_boundary_line(grid, rint(k*X-1), rint((1-k)*Y), rint(k*X-1),     rint(Y-1), FLAG_BOUND_DIR, 1); /* ver */
		/* mark neumann boundary */
		mark_boundary_line(grid,         0,   1,   0, rint((1-k)*Y-1), FLAG_BOUND_NEU, 0);
		mark_boundary_line(grid, rint(k*X), Y-1, X-2,             Y-1, FLAG_BOUND_NEU, 0);
		/* mark outer points */
		recursive_fill_outer(grid, 0, Y-1, 1);
		break;

		case 2:
		/* one conductor */
		k = 0.15;
		mark_boundary_rectangle(grid,         0,         0,             X-1,             Y-1, FLAG_BOUND_DIR, 0);
		mark_boundary_rectangle(grid, rint(k*X), rint(k*Y), rint((1-k)*X-1), rint((1-k)*Y-1), FLAG_BOUND_DIR, 1);
		recursive_fill_outer(grid, X/2, Y/2, 1);
		break;

		case 3:
		/* two square conductors */
		k = 0.3;
		mark_boundary_rectangle(grid, 0, 0, X-1, Y-1, FLAG_BOUND_DIR, 0);
		mark_boundary_square(grid, X/2,     rint(k*Y), X/10, FLAG_BOUND_DIR, 1);
		mark_boundary_square(grid, X/2, rint((1-k)*Y), X/10, FLAG_BOUND_DIR, -1);
		recursive_fill_outer(grid, X/2,     rint(k*Y), 1);
		recursive_fill_outer(grid, X/2, rint((1-k)*Y), -1);
		break;

		case 4:
		/* four square conductors */
		k = 0.2;
		mark_boundary_rectangle(grid, 0, 0, X-1, Y-1, FLAG_BOUND_DIR, 0);
		mark_boundary_square(grid, X/2+k*X, Y/2+k*Y,  X/15, FLAG_BOUND_DIR, -1);
		mark_boundary_square(grid, X/2+k*X, Y/2-k*Y,  X/15, FLAG_BOUND_DIR, -1);
		mark_boundary_square(grid, X/2-k*X, Y/2+k*Y,  X/15, FLAG_BOUND_DIR, 1);
		mark_boundary_square(grid, X/2-k*X, Y/2-k*Y,  X/15, FLAG_BOUND_DIR, 1);
		recursive_fill_outer(grid, X/2+k*X, Y/2+k*Y, -1);
		recursive_fill_outer(grid, X/2+k*X, Y/2-k*Y, -1);
		recursive_fill_outer(grid, X/2-k*X, Y/2+k*Y, 1);
		recursive_fill_outer(grid, X/2-k*X, Y/2-k*Y, 1);
		break;

		case 5:
		/* capacitor */
		k = 0.1;
		mark_boundary_rectangle(grid, 0, 0, X-1, Y-1, FLAG_BOUND_DIR, 0);
		mark_boundary_line(grid, X/2-k*X, Y/2-Y/10, X/2-k*X, Y/2+Y/10, FLAG_BOUND_DIR, -1);
		mark_boundary_line(grid, X/2+k*X, Y/2-Y/10, X/2+k*X, Y/2+Y/10, FLAG_BOUND_DIR, 1);
		//mark_boundary_rectangle(grid, X/2-k*X-1, Y/2-k*Y, X/2-k*X+1, Y/2+k*Y, FLAG_BOUND_DIR, -1);
		//mark_boundary_rectangle(grid, X/2+k*X-1, Y/2-k*Y, X/2+k*X+1, Y/2+k*Y, FLAG_BOUND_DIR, 1);
		recursive_fill_outer(grid, X/2-k*X, Y/2, -1);
		recursive_fill_outer(grid, X/2+k*X, Y/2, 1);
		break;

		case 6:
		/* circolar coaxial cable */
		k = 0.1;
		mark_boundary_circle(grid, X/2, Y/2,  rint(4*k*X), FLAG_BOUND_DIR, 0);
		mark_boundary_circle(grid, X/2, Y/2,    rint(k*X), FLAG_BOUND_DIR, 1);
		recursive_fill_outer(grid, 0,     0, 0);
		recursive_fill_outer(grid, X/2, Y/2, 1);
		break;
	}

	ninner = 0;
	nouter = 0;
	nbound = 0;
	nbound_dir = 0;
	nbound_neu = 0;

	/* mark inner points: all non-outer, non-boundary points are inner.
	   Also, count all points. */
	for (j=0; j<Y; j++) {
		for (i=0; i<X; i++) {
			pnode = grid+ind(i,j);
			if (pnode->flags == 0) {
				pnode->flags |= FLAG_INNER;
				ninner++;
				continue;
			}
			if (pnode->flags & FLAG_BOUND) {
				if (pnode->flags & FLAG_BOUND_DIR)
					nbound_dir++;
				if (pnode->flags & FLAG_BOUND_NEU)
					nbound_neu++;
				continue;
			}
			if (pnode->flags & FLAG_OUTER) {
				//pnode->v = 0;
				nouter++;
			}
		}
	}
	nbound = nbound_dir + nbound_neu;

	printf("Problem: %d\n", problem);
	printf("Output file: %s\n", filename);
	printf("Grid: %dx%d, Total nodes: %d, Mem: %d bytes\n", X, Y, X*Y, X*Y * (int)sizeof(struct node));
	printf("Inner: %d, Outer: %d, Boundary: %d [Dir: %d / Neu: %d].\n\n",
						ninner, nouter, nbound, nbound_dir, nbound_neu);

	if (X*Y != ninner + nouter + nbound) {
		printf("Error: count mismatch\n");
		exit(1);
	}

	/* print the grid/domain */
	printgrid(grid);

	/* allocate memory for the system of equations */
	N = ninner + nbound_neu;
	uind = malloc(N * sizeof(int));
	AB   = malloc(N * sizeof(struct equation));
	Xold = malloc(N * sizeof(double));
	Xnew = malloc(N * sizeof(double));
	if (!uind || !AB || !Xold || !Xnew) {
		printf("Error on malloc()\n");
		exit(1);
	}
	printf("\nTotal unknowns: %d, Mem: %d bytes\n", N, N*(int)(sizeof(int) + sizeof(struct equation) + 2*sizeof(double)) );

	/* clean up memory */
	for (ii=0; ii<N; ii++) {
		*(Xold+ii) = 0.0;
		*(Xnew+ii) = 0.0;
		(AB+ii)->b = 0.0;
		for (j=0; j<5; j++) {
			(AB+ii)->coeff[j] = 0;
			(AB+ii)->index[j] = 0;
		}
	}

	/* prepare uind (unknown indices) vector */
	ii = 0;
	for (i=0; i<X*Y; i++) {
		pnode = grid+i;
		if (pnode->flags & FLAG_INNER || pnode->flags & FLAG_BOUND_NEU) {
			pnode->uindex = ii;
			*(uind+ii) = i;
			ii++;
		}
	}

	/* prepare system of equations */
	for (ii=0; ii<N; ii++) {
		pnode = grid+*(uind+ii);
		peq   = AB+ii;
		//printf("%d_%d_(%02d,%02d)\n", pnode->uindex, pnode->index, pnode->x, pnode->y);

		nn = north(pnode);
		ns = south(pnode);
		ne =  east(pnode);
		nw =  west(pnode);

		peq->coeff[0] = 4;
		peq->index[0] = ii;

		if (pnode->flags & FLAG_BOUND_NEU) {
			//if (nn == -1 || ns == -1 || ne == -1 || nw == -1)
				//printf(" missing node: N-%d: (n%d,s%d,e%d,w%d)\n", ii, nn, ns, ne, nw);
			
			/* north */
			if (nn != -1) {
				pn = grid+nn;
				if (pn->flags & FLAG_BOUND_DIR)
					peq->b += pn->bv;
				if (pn->flags & FLAG_INNER || pn->flags & FLAG_BOUND_NEU) {
					peq->coeff[1] = -1;
					peq->index[1] = pn->uindex;
				}
			} else {
				pn = grid+ns;
				if (pn->flags & FLAG_INNER || pn->flags & FLAG_BOUND_NEU)
					peq->coeff[2] += -1;
			}

			/* south */
			if (ns != -1) {
				pn = grid+ns;
				if (pn->flags & FLAG_BOUND_DIR)
					peq->b += pn->bv;
				if (pn->flags & FLAG_INNER || pn->flags & FLAG_BOUND_NEU) {
					peq->coeff[2] = -1;
					peq->index[2] = pn->uindex;
				}
			} else {
				pn = grid+nn;
				if (pn->flags & FLAG_INNER || pn->flags & FLAG_BOUND_NEU)
					peq->coeff[1] += -1;
			}

			/* east */
			if (ne != -1) {
				pn = grid+ne;
				if (pn->flags & FLAG_BOUND_DIR)
					peq->b += pn->bv;
				if (pn->flags & FLAG_INNER || pn->flags & FLAG_BOUND_NEU) {
					peq->coeff[3] = -1;
					peq->index[3] = pn->uindex;
				}
			} else {
				pn = grid+nw;
				if (pn->flags & FLAG_INNER || pn->flags & FLAG_BOUND_NEU)
					peq->coeff[4] += -1;
			}

			/* west */
			if (nw != -1) {
				pn = grid+nw;
				if (pn->flags & FLAG_BOUND_DIR)
					peq->b += pn->bv;
				if (pn->flags & FLAG_INNER || pn->flags & FLAG_BOUND_NEU) {
					peq->coeff[4] = -1;
					peq->index[4] = pn->uindex;
				}
			} else {
				pn = grid+ne;
				if (pn->flags & FLAG_INNER || pn->flags & FLAG_BOUND_NEU)
					peq->coeff[3] += -1;
			}

		}
		else { /* INNER */
			if (nn == -1 || ns == -1 || ne == -1 || nw == -1)
				printf(" missing node: I-%d: (n%d,s%d,e%d,w%d)\n", ii, nn, ns, ne, nw);

			/* north */
			pn = grid+nn;
			if (pn->flags & FLAG_BOUND_DIR)
				peq->b += pn->bv;
			if (pn->flags & FLAG_INNER || pn->flags & FLAG_BOUND_NEU) {
				peq->coeff[1] = -1;
				peq->index[1] = pn->uindex;
			}

			/* south */
			pn = grid+ns;
			if (pn->flags & FLAG_BOUND_DIR)
				peq->b += pn->bv;
			if (pn->flags & FLAG_INNER || pn->flags & FLAG_BOUND_NEU) {
				peq->coeff[2] = -1;
				peq->index[2] = pn->uindex;
			}

			/* east */
			pn = grid+ne;
			if (pn->flags & FLAG_BOUND_DIR)
				peq->b += pn->bv;
			if (pn->flags & FLAG_INNER || pn->flags & FLAG_BOUND_NEU) {
				peq->coeff[3] = -1;
				peq->index[3] = pn->uindex;
			}

			/* west */
			pn = grid+nw;
			if (pn->flags & FLAG_BOUND_DIR)
				peq->b += pn->bv;
			if (pn->flags & FLAG_INNER || pn->flags & FLAG_BOUND_NEU) {
				peq->coeff[4] = -1;
				peq->index[4] = pn->uindex;
			}
		}
#if 0
		/* print system of equations */
		printf(" Eq: ");
		for (j=0; j<4; j++)
			printf("%d*v_%d + ", peq->coeff[j], peq->index[j]);
		printf("%d*v_%d = %.1f\n", peq->coeff[j], peq->index[j], peq->b);
#endif
	}


	/* solve the system of equations using the Jacobi method */
	printf("Solving the system of equations...\n");
	for (j=0; j<30000; j++) {
		for (ii=0; ii<N; ii++) {
			peq = AB+ii;
			*(Xnew+ii) = (peq->b -  peq->coeff[1] * *(Xold+peq->index[1]) -
									peq->coeff[2] * *(Xold+peq->index[2]) -
									peq->coeff[3] * *(Xold+peq->index[3]) -
									peq->coeff[4] * *(Xold+peq->index[4]) ) / peq->coeff[0];
		}
		/* check if we have reached the requested precision */
		if (!(j%200)) {
			rmsnew = 0.0;
			rms    = 0.0;
			for (ii=0; ii<N; ii++) {
				rmsnew += *(Xnew+ii) * *(Xnew+ii);
//				rms    += fabsf(*(Xnew+ii) * *(Xnew+ii) - *(Xold+ii) * *(Xold+ii));
				rms    += (*(Xnew+ii) - *(Xold+ii)) * (*(Xnew+ii) - *(Xold+ii));
			}
			rms = sqrtf(rms/rmsnew) * 100.0;
			printf(" j=%d\terr=%.4f%%\n", j, rms);
			if (rms < 0.0001) {
				printf(" Solution found!\n");
				break;
			}
		}
		/* swap Xold and Xnew */
		aux  = Xold;
		Xold = Xnew;
		Xnew = aux;
	}

	/* load final values onto grid */
	for (ii=0; ii<N; ii++) {
		pnode = grid+*(uind+ii);
		pnode->v = *(Xnew+ii);
	}

	/* compute normalized capacitance (epsilon = 1, deltaV^2 = 1) */
	/* C = (epsilon / deltaV^2) * int( |grad(v)|^2 dx dy ) [F/m]*/
	printf("Calculating int( |grad(v)|^2 dx dy )\n");
	cnorm = 0.0;
	for (j=0; j<Y; j++) {
		for (i=0; i<X; i++) {
			pnode = grid+ind(i,j);
			if (!(pnode->flags & FLAG_OUTER)) {
				/* first half */
				ns = south(pnode);
				ne =  east(pnode);
				if (ns != -1 && ne != -1) {
					if(!((grid+ns)->flags & FLAG_OUTER) && !((grid+ne)->flags & FLAG_OUTER)) {
						dx = (grid+ne)->v - pnode->v;
						dy = (grid+ns)->v - pnode->v;
						cnorm += 0.5 * (dx*dx + dy*dy);
					}
				}

				/* second half */
				nn = north(pnode);
				nw =  west(pnode);
				if (nn != -1 && nw != -1) {
					if (!((grid+nn)->flags & FLAG_OUTER) && !((grid+nw)->flags & FLAG_OUTER)) {
						dx = pnode->v - (grid+nw)->v;
						dy = pnode->v - (grid+nn)->v;
						cnorm += 0.5 * (dx*dx + dy*dy);
					}
				}
			}
		}
	}
	printf("  int( |grad(v)|^2 dx dy ) = %f\n", cnorm);
	printf("  x2 = %f\n", 2*cnorm);
	printf("  x4 = %f\n", 4*cnorm);

	/* print the whole matrix of values to outfile */
	for (i=0; i<X*Y; i++) {
		fprintf(fp, "%.15f ", (grid+i)->v);
		if ((grid+i)->x == X-1)
			fprintf(fp, "\n");
	}
	fclose(fp);

	return 0;
}
